#
# Copyright 2020 Vianai Systems, Inc. Vianai Confidential & Proprietary Information
#
# Using SHAP() as facility to generate explainable insight of trainig & testing against model
# https://shap.readthedocs.io/en/latest/ 

import shap
import pandas as pd
import xgboost
import lightgbm

class ml_explainer(object):
    """ 
    To generate explainable SHAP value for model at model.py
    """
    def __init__(self, model, desc):
        self.explainer = ml_explainer.gen_explainer(model)
        # for versioning purpose
        self.desc = desc

    @staticmethod
    def model_type(model):
        """
        model_type support xgboost.core.Booster, lightgbm.basic.Booster for now
        """
        if model.__class__ in [xgboost.core.Booster, lightgbm.basic.Booster]:
            return 'Tree'
        else:
            print(model.__class__)
            return 'Unknown'
    # generate explainer based on the model
    @staticmethod
    def gen_explainer(model):
        """
        generate SHAP explainer from model, support Tree only for now
        """
        if ml_explainer.model_type(model) == 'Tree':
            return shap.TreeExplainer(model, feature_perturbation='tree_path_dependent')
        else:
            assert False, 'Model not supported'
        return False

    def generate_shap_value(self, data):
        return self.explainer.shap_values(data)

    # generate shap view as format of:
    # feature, value, shap_value
    def generate_feature_shap(self, data, cols, index_col, extra_cols=[]):
        """
        generate dataframe for features at format of:
        feature, value, shap_value

        Parameters
        ----------
        data: dataframe of training or testing data set
        cols: list of column(str) as list features
        index_col: index column
        extra_cols: additional columns not in cols

        Returns
        -------
        DataFrame: dataframe of the feature value & SAHP
        """
        shap_data = pd.DataFrame(self.generate_shap_value(data[cols]), columns=cols, index=data.index)
        index = 0
        fets = []
        for index in range(0, data.shape[0]):
            df_i = pd.DataFrame({
                                'feature':cols,
                                'value':data.iloc[index][cols].values,
                                'shap_value':shap_data.iloc[index][cols].values
                            })
            if index_col == '':
                df_i['index_col'] = index
            else:
                df_i[index_col] = shap_data.iloc[index][index_col]

            for c in extra_cols:
                df_i[c] = df_fet_v[c]

            df_i['description'] = self.desc
            fets.append(df_i)
        df_fets = pd.concat(fets, axis=0, ignore_index=True)  
        return df_fets



if __name__ == '__main__':
    import xgboost
    X,y = shap.datasets.boston()
    model = xgboost.train({"learning_rate": 0.01}, xgboost.DMatrix(X, label=y), 100)

    # explain the model's predictions using SHAP
    # (same syntax works for LightGBM, CatBoost, scikit-learn and spark models)
    ml_exp = ml_explainer(model, 'test')
    # import pdb; pdb.set_trace()
    print(ml_exp.generate_feature_shap(X, X.columns, ''))