#
# Copyright 2020 Vianai Systems, Inc. Vianai Confidential & Proprietary Information
#

import os, joblib
import pickle
import tensorflow as tf
import tensorflow
import tensorflow.keras.backend as K
from tensorflow.keras.models import load_model
from sklearn.linear_model import LogisticRegression
from tensorflow.keras.callbacks import EarlyStopping
import tensorflow.keras.losses
import numpy as np
from sklearn.model_selection import RepeatedStratifiedKFold
from imblearn.metrics import geometric_mean_score
from sklearn.metrics import f1_score
import string
import spacy
import pandas as pd
from sklearn.metrics import roc_curve, precision_recall_curve, f1_score

class model(object):
    """ Basic model structure.
    """    
    def __init__(self, model_repo, version, **parameter_list):
        self.model_repo = model_repo
        self.version = version
        self.model_persistency_prepare()
        self.estimator = None
        self.setup_pipeline()

    def setup_pipeline(self):
        self.pipeline_name = 'pipeline.pkl'

    # def build_estimator(self, **parameter_list):
    #     raise NotImplementedError

    def save(self, **parameter_list):
        """
        save model pickle file to local
        """
        try:
            joblib.dump(self.estimator, '%s/%s'%(self.model_dir, self.pipeline_name))
        except:
            self.estimator.model.save('%s/model.h5'%self.model_dir)
            self.estimator.model = None   
            joblib.dump(self.estimator, '%s/%s'%(self.model_dir, self.pipeline_name))

    def load(self, **parameter_list):
        """
        read model pickle file from local
        """
        assert os.path.exists(self.model_dir), 'Model directory %s could not located'%self.model_dir
        print(self.model_dir)
        self.estimator = joblib.load('%s/%s'%(self.model_dir, self.pipeline_name))

    def set_estimator(self, est):
        self.estimator = est

    def train(self, X_train, y_train, **parameter_list):
        """
        Train model

        Parameters
        ----------
        X_train: 
            training features
        y_train: 
            training target
        parameter_list: dict
            dictionary of hyper-parameters
        """
        # self.estimator = self.build_estimator(parameter_list)
        assert self.estimator is not None, 'estimator not setup yet'
        self.estimator.fit(X_train, y_train)

    def predict_prob(self, X_test):
        raise NotImplementedError

    def predict(self, X_test):
        """
        model prediction

        Parameters
        ----------
        X_test: test features

        Returns
        -------
        y_pred: 
            prediction
        """
        return self.estimator.predict(X_test)

    def model_persistency_prepare(self):
        self.work_dir = '%s/%s'%(self.model_repo, self.version)
        self.model_dir = '%s/model'%self.work_dir
        self.test_dir = '%s/test'%self.work_dir

        for _dir in [self.work_dir, self.model_dir, self.test_dir]:
            if not os.path.exists(_dir): 
                os.makedirs(_dir)

# TODO: to update for arima model
class arimaModel(model):
    """
    Build Arima Model
    """
    def __init__(self, model_repo, version,term_id, **parameter_list):
        self.model_repo = model_repo
        self.version = version
        self.model_persistency_prepare()
        self.estimator = None
        self.setup_pipeline()
        self.term_id = term_id

    

    def save(self, **parameter_list):
        joblib.dump(self.estimator, '%s/%s_%s'%(self.model_dir, self.term_id, self.pipeline_name))

    def load(self, **parameter_list):
        assert os.path.exists(self.model_dir), 'Model directory %s could not located'%self.model_dir
        self.estimator = joblib.load('%s/%s_%s'%(self.model_dir, self.term_id, self.pipeline_name))

    def train(self, X_train, y_train, **parameter_list):
        assert self.estimator is not None, 'estimator not setup yet'
        # attention, the y and x switched by the arima model
        X_train_processed = self.estimator.get_params()['preprocessor'].fit_transform(X_train)
        #print('the shape of train' , X_train_processed.shape)
        self.estimator.get_params()['ml_model'].fit(y_train, X_train_processed)
        #self.estimator.fit(y_train, X_train)

    def update(self,new_y,new_x):
        self.estimator.get_params()['ml_model'].update(new_y,new_x)

    def forecast(self, n_periods, exogenous):
        pred = self.estimator.get_params()['ml_model'].predict(n_periods=n_periods,
                                                            exogenous=exogenous,
                                                            return_conf_int=False)
        pred = np.where(pred < 0, 0, pred)
        return pred

class regularModel(model):
    """ Regular Regression Model
    """
    def train(self, X_train, y_train, **parameter_list):
        """
        Train model

        Parameters
        ----------
        X_train: 
            training features
        y_train: 
            training target
        parameter_list: dict
            dictionary of hyper-parameters
        """
        #X_train_processed = self.estimator.get_params()['preprocessor'].fit_transform(X_train)
        #self.estimator.get_params()['ml_model'].fit(X_train_processed, y_train)
        self.estimator.fit(X_train, y_train)
        
    def predict(self, test_data, **parameter_list):
        """
        model prediction

        Parameters
        ----------
        test_data: 
            test features
        parameter_list: dict
            dictionary of hyper-parameters

        Returns
        -------
        y_pred: 
            prediction
        """
        #test_data_processed = self.estimator.get_params()['preprocessor'].transform(test_data)
        #y_pred = self.estimator.get_params()['ml_model'].predict(test_data_processed)
        y_pred = self.estimator.predict(test_data)
        #y_pred = np.where(y_pred < 0, 0, y_pred)
        return y_pred
    

class NLPClassificationModel(model):
    def __init__(self, model_repo, version, **parameter_list):
        self.best_thresh = 0.5
        self.model_repo = model_repo
        self.version = version
        self.parameter_list = parameter_list
        self.criterion = 'criterion' in parameter_list.keys() and parameter_list['criterion']
        self.work_dir = '%s/%s'%(self.model_repo, self.version)
        self.model_dir = '%s/model'%self.work_dir
        self.test_dir = '%s/test'%self.work_dir
        self.pipeline_name = 'pipeline.pkl'
        for _dir in [self.work_dir, self.model_dir, self.test_dir]:
            if not os.path.exists(_dir): 
                os.makedirs(_dir)
                              
    def cleanup_text(self, train_x, logging=False):
        from nltk.corpus import stopwords
        stopwords = stopwords.words('english')
        punctuations = string.punctuation
        docs = [text for text in train_x]
        
        punctuations = string.punctuation
        texts = []
        nlp = spacy.load('en_core_web_sm')
        
        
        for doc in docs:
            doc = nlp(doc, disable=['parser', 'ner'])
            tokens = [tok.lemma_.lower().strip() for tok in doc if tok.lemma_ != '-PRON-']
            tokens = [tok for tok in tokens if tok not in stopwords and tok not in punctuations]
            tokens = ' '.join(tokens)
            texts.append(tokens)
        return pd.Series(texts)
    
      
    def vectorize(self, clean_text):
        train_vec = []
        nlp = spacy.load('en_core_web_sm')
        train_vec = [doc.vector for doc in nlp.pipe(clean_text)]
        train_vec = np.array(train_vec)
        return train_vec
    
    def threshold_search(self, y_true, y_proba, criterion):
        if criterion =='f1':
            precision , recall, thresholds = precision_recall_curve(y_true, y_proba)
            thresholds = np.append(thresholds, 1.001) 
            F = 2 / (1/precision + 1/recall)
            best_score = np.max(F)
            return thresholds[np.argmax(F)]

        elif criterion == 'auc':
            fpr, tpr, thresholds = roc_curve(y_true, y_proba)
            # calculate the g-mean for each threshold
            thresholds = np.append(thresholds, 1.001)
            gmeans = np.sqrt(tpr * (1-fpr))
            # locate the index of the largest g-mean
            return thresholds[np.argmax(gmeans)]
        return
        
    def scoring(self, y_true, y_proba, criterion, verbose=True):
        rkf = RepeatedStratifiedKFold(n_splits=5, n_repeats=10)
        scores = []
        ths = []
        for train_index, test_index in rkf.split(y_true, y_true):
            y_prob_train, y_prob_test = y_proba[train_index], y_proba[test_index]
            y_true_train, y_true_test = y_true[train_index], y_true[test_index]

            # determine best threshold on 'train' part 
            best_threshold = self.threshold_search(y_true_train, y_prob_train, criterion)

            if criterion == 'f1':
                # use this threshold on 'test' part for score 
                scores.append(f1_score(y_true_test, (y_prob_test >= best_threshold).astype(int)))
            elif criterion == 'auc':
                scores.append(geometric_mean_score(y_true_test, (y_prob_test >= best_threshold).astype(int)))
            ths.append(best_threshold)

        best_thre = np.mean(ths)
        score = np.mean(scores)

        if verbose: print(f'Best threshold: {np.round(best_thre, 4)}, Score: {np.round(score,5)}')

        return best_thre, score
    
    
    def get_threshold(self, X_train, y_train, criterion):
        self.estimator.fit(X_train, y_train)
        y_proba = self.estimator.predict_proba(X_train)[:,1]
        print(y_proba, y_proba.shape)
        best_thresh, score = self.scoring(y_train, y_proba, criterion, verbose=True)
        return best_thresh
   
    
    def train(self, X_train, y_train, **parameter_list):
        clean_text = self.cleanup_text(X_train)
        X_train = self.vectorize(clean_text)
        self.best_thresh = self.get_threshold(X_train, y_train, self.criterion)
        self.estimator.fit(X_train, y_train)

    def predict(self, test_data, **parameter_list):
        print(type(self.estimator), self.estimator)
        if tensorflow.keras.wrappers.scikit_learn.KerasClassifier == type(self.estimator):
            self.estimator = joblib.load('%s/pipeline.pkl'%(self.model_dir))
            self.estimator.model = load_model('%s/model.h5'%(self.model_dir))
        else:
            self.load(**parameter_list)
       
        clean_text = self.cleanup_text(test_data)
        test_data = self.vectorize(clean_text)
        y_proba = self.estimator.predict_proba(test_data)[:,1]
        y_pred = (y_proba >= self.best_thresh).astype(int)
        print(y_pred)
        return y_pred
        

class kerasModel(model):
    def save(self, **parameter_list):
        """save model to local"""
        assert tf.keras.wrappers.scikit_learn.KerasClassifier == type(self.estimator.named_steps['ml_model']) or  \
               tf.keras.wrappers.scikit_learn.KerasRegressor == type(self.estimator.named_steps['ml_model']) \
                        , 'module %s is not keras ml_model'%type(self.estimator.named_steps['ml_model'])
            
        self.estimator.named_steps['ml_model'].model.save('%s/model.h5'%self.model_dir)
        self.estimator.named_steps['ml_model'].model = None   

        joblib.dump(self.estimator, '%s/%s'%(self.model_dir, self.pipeline_name))
 
    def load(self, **parameter_list):
        """read model from local"""
        def myloss(y_true, y_pred):
            y_true = tf.cast(y_true,tf.float32)
            y_pred = tf.cast(y_pred,tf.float32)
            idx  = tf.not_equal(y_true,0)
            y_true = tf.boolean_mask(y_true, idx)
            y_pred = tf.boolean_mask(y_pred, idx)
            idx  = tf.not_equal(y_pred,0)
            y_true = tf.boolean_mask(y_true, idx)
            y_pred = tf.boolean_mask(y_pred, idx)
            loss = tf.cond(tf.equal(tf.shape(y_pred)[0], 0), lambda: tf.constant(0, dtype=tf.float32), lambda: tf.keras.backend.sum(tf.abs(y_true-y_pred))/tf.keras.backend.sum((y_true+y_pred)))
            return loss 

        # keras.losses.myloss2 = myloss2
        self.estimator = joblib.load('%s/pipeline.pkl'%(self.model_dir))
        self.estimator.named_steps['ml_model'].model = load_model('%s/model.h5'%(self.model_dir))


class FCModel(kerasModel):
    def __init__(self, model_repo, version, **parameter_list):
        self.model_repo = model_repo
        self.version = version
        self.parameter_list = parameter_list
        self.patience = 'patience' in parameter_list.keys() and parameter_list['patience'] or 3
        self.batch_size = 'batch_size' in parameter_list.keys() and parameter_list['batch_size'] or 32
        self.work_dir = '%s/%s'%(self.model_repo, self.version)
        self.model_dir = '%s/model'%self.work_dir
        self.test_dir = '%s/test'%self.work_dir
        self.pipeline_name = 'pipeline.pkl'
        
        for _dir in [self.work_dir, self.model_dir, self.test_dir]:
            if not os.path.exists(_dir): 
                os.makedirs(_dir)
    def train(self, X_train, y_train,X_validation,y_validation, **parameter_list):
        """
        Train model

        Parameters
        ----------
        X_train: 
            training features
        y_train: 
            training target
        X_validation: 
            validation features
        y_validation: 
            validation target
        parameter_list: dict
            dictionary of hyper-parameters

        """
        X_train_processed = self.estimator.get_params()['preprocessor'].fit_transform(X_train)
        try:
            X_train_processed = X_train_processed.todense()
        except:
            X_train_processed = X_train_processed 
        X_validation_processed = self.estimator.get_params()['preprocessor'].transform(X_validation)
        try:
            X_validation_processed = X_validation_processed.todense()
        except:
            X_validation_processed = X_validation_processed 
       
        self.estimator.get_params()['ml_model'].fit(X_train_processed, y_train,validation_data = (X_validation_processed,y_validation), 
        callbacks=[EarlyStopping(monitor='val_loss', patience=self.patience)],batch_size = self.batch_size,  **parameter_list)
        
    def predict(self, test_data, **parameter_list):
        """
        model prediction

        Parameters
        ----------
        test_data: 
            test features
        parameter_list: dict
            dictionary of hyper-parameters

        Returns
        -------
        y_pred: 
            prediction
        """
        self.load(**parameter_list)
        test_data_processed = self.estimator.get_params()['preprocessor'].transform(test_data)
        try:
            test_data_processed = test_data_processed.todense()
        except:
            test_data_processed = test_data_processed 
        y_pred = self.estimator.get_params()['ml_model'].predict(test_data_processed, verbose=0)
        return y_pred.flatten()


