import yaml
import psycopg2
from sqlalchemy import create_engine
import pandas as pd 
import numpy as np
from ml_experiments.experiment import * 
from ml_experiments.model import *
from tensorflow.keras.wrappers.scikit_learn import KerasClassifier


with open("config.yaml", "r") as ymlfile:
    config = yaml.load(ymlfile, Loader=yaml.FullLoader)


def delete_all_tables():
    # print ('========delete all tables==========')
    username = config['Data']['DataBase']['username']
    password = config['Data']['DataBase']['password']
    databasename = config['Data']['DataBase']['databasename']

    conn = psycopg2.connect("dbname={} user={}".format(databasename, username))
    cur = conn.cursor()
    cur.execute("DROP SCHEMA public CASCADE;")
    cur.execute("CREATE SCHEMA public;")
    conn.commit()

    cur.close()
    conn.close()


def upload_to_sql(df, tb_name,if_exists='append'):
    username = config['Data']['DataBase']['username']
    password = config['Data']['DataBase']['password']
    databasename = config['Data']['DataBase']['databasename']
    engine = create_engine('postgresql://{}:{}@localhost/{}'.format(username,
                                                                    password,
                                                                    databasename))
    df.to_sql(tb_name, engine, if_exists=if_exists, index=False)


def get_config(cat_cols, cont_cols, params, train_df, valid_df, target,
               df_test, encode_method, desc, ProblemType, ModelType):
    config_good = config['MLModel'][ProblemType][ModelType]['config_good']
    config_good['estimator_param'] = params
    config_good['train'] = train_df
    config_good['validation'] = valid_df
    config_good['test'] = df_test
    config_good['cat_cols'] = cat_cols
    config_good['cont_cols'] = cont_cols
    config_good['encode_method'] = encode_method
    config_good['desc'] = desc
    config_good['target'] = target

    return config_good


def get_experiment_df(cat_cols, cont_cols, params,desc):
    cols = ','.join(cont_cols+cat_cols)
    para = ''
    for pa in params:
        if 'ml_model' in pa:
            para = para + pa.split('__')[1]+':'+str(params[pa])+','
    return cols, para

def build_model(n_features=100):
    model = Sequential()
    # Densely Connected Neural Network (Multi-Layer Perceptron)
    model.add(Dense(512, activation='relu', kernel_initializer='he_normal', input_dim=n_features))
    model.add(Dropout(0.2))
    model.add(Dense(256, activation='relu', kernel_initializer='he_normal'))
    model.add(Dropout(0.2))
    model.add(Dense(150, activation='relu', kernel_initializer='he_normal'))
    model.add(Dropout(0.2))
    model.add(Dense(50, activation='relu', kernel_initializer='he_normal'))
    model.add(Dropout(0.2))
    model.add(Dense(2, activation='sigmoid'))
    model.compile(optimizer='adam', loss='sparse_categorical_crossentropy', metrics=['acc'])
    return model


def main():
    if config['Data']['DataBase']['CleanDataBase']:
        delete_all_tables()

    ProblemType = config['MLModel']['ProblemType']
    ModelType = config['MLModel']['ModelType']
    dataname = config['Data']['DataName']
    cat_cols = config['MLModel'][ProblemType]['cat_cols']
    cont_cols = config['MLModel'][ProblemType]['cont_cols']
    dt_col = config['Data']['dt_col']
    index = config['Data']['index']
    target = config['Data']['target']
    group = config['Data']['group']
    split_time = config['Data']['DataSplit']['split_time']
    test_size = float(config['Data']['DataSplit']['test_size'])
    generate_shap = config['MLModel']['generate_shap']
    encode_method = config['MLModel'][ProblemType][ModelType]['config_good']['encode_method']

    #shap_params = get_shap_params(ProblemType)

    desc = config['MLModel']['desc']
    params = config['MLModel'][ProblemType][ModelType]['params']

    from_local = config['Data']['FromLocal']
    
    df = load_data(dataname, from_local)
    df_viz = df.copy()

    if ProblemType == 'Regression':
        df_viz = df_viz.rename(columns={dt_col: 'dt_col',
                                        target: 'target',
                                        group: 'group'})

        df_viz['dt_col'] = pd.to_datetime(df_viz['dt_col'])
        df_viz['target'] = df_viz['target'].astype('f')
        df_viz['group'] = df_viz['group'].astype('str')

        upload_to_sql(df_viz, 'data_viz_rg', if_exists='replace')

        df_train, df_test = data_split(df, method='time',
                                        split_time=split_time,
                                        dt_col=dt_col)

        train_df, valid_df = data_split(df_train, method='random',
                                       test_size=test_size,
                                       random_state=42)
        
        if ModelType == 'Lgbm':
            config_good = get_config(cat_cols, cont_cols, params, train_df, valid_df, target,
                                     df_test, encode_method, desc, ProblemType, ModelType)
            config_good['estimator'] = build_estimator(cat_cols, cont_cols,
                                                lightgbm.LGBMRegressor(),encode_method)

            rgExperiment = regressionExperiment(**config_good)
            rgExperiment.run_experiment()

            print("=========upload shap values=======")
            rgExperiment.train_shap()
            test_shap_df = rgExperiment.generate_shap_test()
            test_shap_df['description'] = desc
            test_shap_df = test_shap_df.rename(columns={dt_col:'dt_col',group:'group'})
            upload_to_sql(test_shap_df,'test_shap_values_rg')
        
        elif ModelType == 'Keras':
            if encode_method == 'onehot':
                n_features = get_n_features(df_train, cont_cols, cat_cols)
            elif encode_method == 'label':
                n_features = len(cat_cols+cont_cols)
            params['ml_model__n_features'] = n_features
            config_good = get_config(cat_cols, cont_cols, params, train_df, valid_df,target,
                                     df_test, encode_method, desc, ProblemType, ModelType)
            fc_reg = tf.keras.wrappers.scikit_learn.KerasRegressor(build_fn = create_fc_model,     
                                                                epochs = 100,                        
                                                                verbose = 1)
            config_good['estimator'] = build_estimator(cat_cols, cont_cols, fc_reg, encode_method)

            fcExperiment = NNRegressionExperiment(**config_good)
            fcExperiment.run_experiment()

        file_test = '%s/%s/test/test.csv'
        file_metrics = '%s/%s/test/metrics.csv'
        df_result = pd.read_csv(file_test%(config_good['model_repo'],
                                        config_good['desc']))
        df_metrics = pd.read_csv(file_metrics%(config_good['model_repo'],
                                            config_good['desc']))
        df_result['description'] = desc
        df_result = df_result.rename(columns={dt_col:'dt_col',group:'group'})
        cols, para = get_experiment_df(cat_cols, cont_cols, params,desc)
        df_result['columns'] = cols
        df_result['params'] = para
        print("========upload test result========")
        upload_to_sql(df_result,'test_result_rg')


    elif ProblemType == 'Classification':
        df_viz = df_viz.rename(columns={index: 'index',
                                        target: 'target'})

        df_viz['target'] = df_viz['target'].astype('str')

        upload_to_sql(df_viz,'data_viz_cls',if_exists='replace')

        df_train, df_test = data_split(df, method='random',
                                       test_size=0.2,
                                       random_state=42)

        train_df, valid_df = data_split(df_train, method='random',
                                       test_size=0.2,
                                       random_state=42)
        
        if ModelType == 'Lgbm':
            config_good = get_config(cat_cols, cont_cols, params, train_df,
                                     valid_df, target, df_test, encode_method, desc,
                                     ProblemType, ModelType)
            config_good['estimator'] = build_estimator(cat_cols, cont_cols,
                                                       lightgbm.LGBMClassifier(), encode_method)

            clfExperiment = classificationExperiment(**config_good)
            clfExperiment.run_experiment()

            print("=========upload shap values=======")
            clfExperiment.train_shap()
            test_shap_df = clfExperiment.generate_shap_test()
            test_shap_df['description'] = desc
            test_shap_df = test_shap_df.rename(columns={index:'index'})
            upload_to_sql(test_shap_df, 'test_shap_values_cls')
        
        elif ModelType == 'Keras':
            if encode_method == 'onehot':
                n_features = get_n_features(df_train, cont_cols, cat_cols)
            elif encode_method == 'label':
                n_features = len(cat_cols+cont_cols)
            params['ml_model__n_features'] = n_features
            config_good = get_config(cat_cols, cont_cols, params, train_df, valid_df, target,
                                     df_test, encode_method, desc, ProblemType, ModelType)
            fc_classifier = KerasClassifier(build_fn=build_model, epochs=10,               
                                            verbose=1, batch_size=10,
                                            callbacks=[EarlyStopping(monitor='loss', patience=5)])
            config_good['estimator'] = build_estimator(cat_cols, cont_cols, fc_classifier,encode_method)

            clfExperiment = NNClassificationExperiment(**config_good)
            clfExperiment.run_experiment()

        file_test = '%s/%s/test/test.csv'
        file_metrics = '%s/%s/test/metrics.csv'
        df_result = pd.read_csv(file_test%(config_good['model_repo'],
                                        config_good['desc']))
        df_metrics = pd.read_csv(file_metrics%(config_good['model_repo'],
                                            config_good['desc']))
        df_result['description'] = desc
        df_result = df_result.rename(columns={index:'index'})
        cols, para = get_experiment_df(cat_cols, cont_cols, params,desc)
        df_result['columns'] = cols
        df_result['params'] = para
        print("========upload test result========")
        upload_to_sql(df_result,'test_result_cls')

    # print("========upload experiment df========")
    # experiment_df = get_experiment_df(cat_cols, cont_cols, params,desc)
    # upload_to_sql(experiment_df, 'experiment_df')


if __name__ == "__main__":
    main()


